<?php 
#
# twiggi message threading (C) 2001 neddix, Stuttgart
#
# based on the document by Jamie Zawinski, www.jwz.org
#
$msg_table=array();
class ThreadContainer {
	var $message;
	var $parent;
	var $child;
	var $next;
	var $uniq;

	Function printContainer ( $title="", $color="#880000" )
		{
		GLOBAL $msg_table;
		ob_start();
		if ( $title )
			{
			print "$title\n";
			}
		print "Container: " . $this->uniq . "\n";
		print "message(". $this->message . "):" . ($msg_table[$this->message]->subject) ."\n"; 
		print "id:" . ($msg_table[$this->message]->id) ."\n"; 
		if ( $msg_table[$this->message]->references )
			{
			print "refs:" . join (":",($msg_table[$this->message]->references)) ."\n"; 
			}
		if ( $this->parent )
			{
			print "parent:(" . $this->parent->uniq . ") " . $msg_table[$this->parent->message]->id ."\n"; 
			}
		if ( $this->child )
			{
			print "child:(" . $this->child->uniq . ") " . $msg_table[$this->child->message]->id ."\n"; 
			}
		if ( $this->next )
			{
			print "next:(" . $this->next->uniq . ") " . $msg_table[$this->next->message]->id ."\n"; 
			}
		$out=ob_get_contents();
		ob_end_clean();
		//ADMIN ++ ADDED (htmlentities -> TWIGhtmlentities)
		print "<pre><font color=$color size=+1>" . TWIGhtmlentities($out) . "</font></pre>";
		}

	Function printThread ( $title="" )
		{
		GLOBAL $msg_table;
		STATIC $nesting=0, $cc;
		$cc++;
		if ( !isset ($this->message)  )
			{
				print "this is a dummy<br>";
			}
		$this->printContainer ( $msg_table, $title . " ($cc)" , "#008800");
		if ( $this->child->uniq )
			{
			$nesting++;
			$this->child->printThread ( $msg_table, "nesting=$nesting" );
			$nesting--;
			}
		if ( $this->next->uniq )
			{
			$this->next->printThread ( $msg_table, "nesting=$nesting" );
			}
		}

	Function flushThread( &$list )
		{
		GLOBAL $msg_table;
		STATIC $nesting=0, $cc=0, $threadid=0, $result=array();
		if ( $nesting==0 )
			{
			$threadid++;
			}
		if ( isset( $this->message ) )
			{
			$result[$cc]->nesting=$nesting;
			$result[$cc]->threadid=$threadid;
			$list[$cc]= $msg_table[$this->message]->imapid;
			$cc++;
		}
		
		if ( $this->child->uniq )
			{
			$nesting++;
			$this->child->flushThread ( $list );
			$nesting--;
			}
		if ( $this->next->uniq )
			{
			return $this->next->flushThread ( $list );
			}
		if ( $nesting == 0 )
			{
			return $result;
			}
		}

	Function isEqual ( $c )
		{
		return $this->uniq == $c->uniq;
		}

	# check if target appears in childrens list 
	Function FindChild ( &$target )
		{
		GLOBAL $msg_table;
		if ( !$this->child->uniq  )
			{
			return FALSE; // has no children
			}
		elseif ( $this->child->isEqual($target) )
			{
			return TRUE; // found!
			}
		else
			{
			return $this->child->FindChild ( $target );
			}
		}

	function ThreadContainer ()
		{
		# $this->message=NULL;
		# $this->parent=NULL;
		# $this->child=NULL;
		# $this->next=NULL;
		$this->uniq=uniqid("", TRUE);
		}
}

	Function newThreadContainer ()
	{
	STATIC $oa=array();
	$n=array_push ( $oa, new ThreadContainer );
	return $oa[$n-1];
	}

#  Walk through the threads and discard any empty container objects.  
Function pruneEmptyContainers ( &$parent )
	{
	GLOBAL $msg_table;
	unset( $prev );
	for( $prev, $container = &$parent->child, $next = &$container->next;
			$container->uniq; 
			$prev = &$container, $container = &$next, $next = !$container->uniq  ? NULL : $next = &$container->next )
		{
		if ( !isset($container->message) && !$container->child->uniq )
			{
			if ( !$prev->uniq )
				{
				$parent->child = &$container->next;
				}
			else
				{
				$prev->next = &$container->next;
				}
			$container = &$prev;
			}
		elseif ( !isset($container->message) &&			// expired, and
					$container->child->uniq &&		// has kids, and
					( $container->parent->uniq ||		// not at root, or 
					!$container->child->next->uniq ) )	// only one kid 
			{
			#  Expired message with kids.  Promote the kids to this level.
			$kids = &$container->child;

			#  Remove this container from the list, replacing it with `kids'.
			if ( !$prev->uniq ) 
				{
				$parent->child = &$kids;
				}
			else
				{
				$prev->next = &$kids;
				}
			#  make each child's parent be this level's parent. 
			#  make the last child's next be this container's next

			for ( $tail = &$kids; $tail->next->uniq; $tail = &$tail->next ) 
				{
				$tail->parent = &$container->parent;
				}
			$tail->parent = &$container->parent;
			$tail->next = &$container->next;

			$next = &$kids; 	
			
			$container = &$prev;
			}
		elseif ( $container->child->uniq )
			{
			#  A real message with kids.
			pruneEmptyContainers( $container ); 
			}

		}
	}

Function reverseChildren( &$parent )
	{
	if ( $parent->child->uniq )
		{
		for( $prev, $kid = &$parent->child, $rest = &$kid->next;
				$kid->uniq;
				$prev = &$kid, $kid = &$rest, $rest = !$rest->uniq  ? NULL : $rest = &$rest->next )
			{
			$kid->next=&$prev;
			}
		$parent->child=&$prev;

		for( $kid=&$parent->child; $kid->uniq; $kid=&$kid->next )
			{
			reverseChildren( $kid );
			}
		}
	}

	#  gather subjects:
	#  If any two members of the root set have the same subject, merge them.

Function gatherSubjects( &$root )
	{
	GLOBAL $msg_table, $reply_regexp;

	$subj_table=array();

	Function isReply ( $subj )
		{
		GLOBAL $reply_regexp;
		return preg_match( "/$reply_regexp/", $subj );
		}

	$cc = 0;
	for ( $c = &$root->child; $c->uniq ; $c = &$c->next)
		{
		$message = &$c->message;
		#  If there is no message, this is a dummy node in the root set.
		if ( !isset( $message ) )
			{
			$message = &$c->child->message;
			}

		$subj = preg_replace ("/$reply_regexp/", "", $msg_table[$message]->subject );
		if ( empty( $subj ) ) { continue; }
		$old = &$subj_table[$subj];


		if ( !$subj_table[$subj] && 
			( !$old->uniq ||
			( !isset( $c->message ) && isset( $old->message ) ) ||
			 ( isset( $old->message ) && isReply( $msg_table[$old->message]->subject ) &&
			   isset( $c->message ) && !isReply( $msg_table[$c->message]->subject ) ) ) )
			{
			$subj_table[$subj] = &$c;
			$cc++;
			}
		}
	#  if the table is empty, we're done.
	if ( $cc == 0  ) { return; }

	unset( $prev );
	for ( $prev, $c = &$root->child, $rest = &$c->next; 
			$c->uniq ; 
			$prev = &$c, $c = &$rest, !$rest->uniq ? NULL : $rest=&$rest->next ) 
		{
		$message = &$c->message;
		if ( !isset( $message ) ) // might be a dummy -- see above 
			{
			$message = &$c->child->message;
			}

		$subj = preg_replace ("/$reply_regexp/", "", $msg_table[$message]->subject );
		#  Don't thread together all subjectless messages; let them dangle.
		if ( !$subj ) 
			{ 
			continue; 
			}

		$old = &$subj_table[$subj];

		if ( $old->isEqual ($c) ) 
			{ 
			continue; 
			} // oops, that's us

		# Don't merge different threads even if they have same subject	
		if ( isset($c->message) && isset($old->message) && 
			!isReply ($msg_table[$c->message]->subject) 
			&& !isReply ($msg_table[$old->message]->subject) 
			&& $c->child->uniq 
			&& $old->child->uniq )
			{
			continue;
			}


		if ( !$prev->uniq )
			{
			$root->child = &$c->next; 
			}
		else
			{
			$prev->next = &$c->next;
			}
		unset ( $c->next );

		if ( !isset( $old->message ) && !isset( $c->message ) ) 
			{  
			#  They're both dummies; merge them. 
			for ( $tail = &$old->child;
					$tail->uniq && $tail->next->uniq; 
					$tail = &$tail->next) {} ;
			$tail->next = &$c->child;
			for ( $tail = &$c->child; $tail->uniq; $tail = &$tail->next ) 
				{
				$tail->parent = &$old;
				}
			unset ($c->child);
			}

		elseif ( !isset( $old->message ) ||									// old is empty, or
					( isset( $c->message ) &&
					isReply( $msg_table[$c->message]->subject ) &&       	//   c has Re, and 
					!isReply( $msg_table[$old->message]->subject ) ) )   	//   old does not. 
			{
			#  Make this message be a child of the other.
			$c->parent = &$old;
			$c->next = &$old->child;
			$old->child = &$c;
			}

		elseif ( isset( $c->message ) &&
				isReply( $msg_table[$old->message]->subject ) &&       	//   old has Re, and 
				!isReply( $msg_table[$c->message]->subject ) )   		//   c does not. 
			{
			#  Make this message be a child of the other.
			$c->parent = &$old;
			$c->next = &$old->child;
			$old->child = &$c;

			$m = $c->message;
			$c->message = $old->message;
			$old->message = $m;

			}
		else
			{
			#  Make the old and new messages be children of a new dummy container. 
			$newc = &newThreadContainer ();
			$newc->message = $old->message;
			$newc->child = &$old->child;

			for ( $tail = &$newc->child; $tail->uniq; $tail = &$tail->next)
				{
				$tail->parent = &$newc;
				}

			unset( $old->message );
			unset( $old->child );
			$c->parent = &$old; 
			$newc->parent = &$old;
			$old->child = &$c;
			$c->next = &$newc;
			}

		#  we've done a merge, so keep the same `prev' next time around.
		$c = &$prev; 
		
		}
	}


Function GetThreadedList ( &$mbox, &$list, &$thread, &$count )
	{
	GLOBAL $config, $dbconfig, $msg_table, $session, $login, $reply_regexp;

	$mbx=TWIGMailGetFolderString( $config["imap_server"], $config["imap_port"], "", $session["mailbox"] );

	if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: imap_num_recent imap_num_msg");
	$Nmsgs=imap_num_msg ($mbox);
	if ($config["tdbg"]) timedebug_stop ($tdbg);

	if ( $config["cachethreshold"] >0 && $config["cachethreshold"] < $Nmsgs )
		{
		$force_reread = ( imap_num_recent ($mbox) != 0 );
		$fingerprint_thread = md5 ( $Nmsgs . serialize( $list ) );

		if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: fingerprint_thread db query\nfp=$fingerprint_thread");
		$query="SELECT cache FROM " . $dbconfig["threadcache"] 
					. " WHERE fingerprint='" . $fingerprint_thread . "'"
					. " AND username='" . $login["username"] . "'"
					. " AND mailbox='" . SQLString( $mbx ) . "'"
					. " AND type='T'";
		$result=dbQuery ( $query );
		$ret=dbResultArray( $result, 0 );
		if ($config["tdbg"]) timedebug_stop ($tdbg);
    
		if ( $ret && file_exists( $ret[0] ) && !$force_reread )
			{
			if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: unserialize");

			$query="UPDATE " . $dbconfig["threadcache"] . " SET thedate=NULL WHERE fingerprint='" . $fingerprint_thread . "'";
			dbQuery ( $query );

			$fp = fopen ( $ret[0], "r");
			while ( !feof( $fp ) )
				{
				$cache .= fgets( $fp, 1048576 ); 
				}
			fclose ($fp);
			$cache = unserialize( $cache );
			$list=$cache["list"];
			$thread=$cache["thread"];
			if ($config["tdbg"]) timedebug_stop ($tdbg);
			return;
			}
		}

	$all=thread_imap_fetch_overview ($mbox, $mbx ) ;

	if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: build the thread");

	$flag=array();
	#  build the message table
 	for( $i = 0; $i < $count; $i++ ) 
		{
		$headers = $all[$list[$i]];
		$m->imapid = $list[$i];
		$m->subject = $headers->subject;
			if ( $flag[$headers->message_id] != 1 ) 
			{
			$m->id = $headers->message_id;
			$flag[$headers->message_id]=1;
			}
		else
			{
			$m->id = "<Bogus-id:" . ($bogus_id_count++) . ":" . uniqid("", TRUE) . ">";
			}
		$revref[$m->id]=$i;
		if ( $headers->references )
			{
			$m->references=preg_split ("/[ ]/", $headers->references);
			}
		elseif ($headers->in_reply_to )
			{
			$m->references=array($headers->in_reply_to);
			}
		else
			{
			$m->references=array();
			}
		$msg_table[]=$m;
		}

 	for( $i = 0; $i < $count; $i++ ) 
		{
		$id = $msg_table[$i]->id;
		$c = &$id_table[$id]; 
		if ( $c->uniq )
			{
			if ( !isset( $c->message ) )
				{
				$msg_table[$i]->id = "<Bogus-id:" . $id . ":" . ($bogus_id_count++) . ":" . uniqid("", TRUE) . ">"; 
				unset ($c);
				}
			else
				{
				$c->message = $i;
				}
			}

		#  Create a ThreadContainer for this message, and index it in the hash table.
		if ( !$c->uniq )
			{
			$c = &newThreadContainer(); 
			$c->message = $i;
			$id_table[$id] = &$c;
			}
		
		#  for each element in the reference field ...
		unset( $parent_ref );
		for ( $k=0, $nr=count( $msg_table[$i]->references ); $k<$nr; $k++ )
			{
			#  find a Container for this ID
			$ref = &$id_table[$msg_table[$i]->references[$k]];
			if ( !$ref->uniq )
				{
				$ref = &newThreadContainer(); 
				#$ref->uniq .= "-" . $msg_table[$i]->references[$k];#debug!!!
				$ref->message=$revref[$msg_table[$i]->references[$k]];
				$id_table[$msg_table[$i]->references[$k]] = &$ref;
				}

			#  Link the References field's Containers together 
			if ( $parent_ref->uniq  && 			// there is a parent
				 !$ref->parent->uniq && 			// don't have a parent already
				 !$parent_ref->isEqual($ref) &&		// not a tight loop 
				 !$ref->FindChild ( $parent_ref ) )	// not a wide loop 
				{
				$ref->parent = &$parent_ref;
				$ref->next = &$parent_ref->child;
				$parent_ref->child = &$ref;
				}
			$parent_ref = &$ref;
			}

		if(	$parent_ref->uniq && 
			( $parent_ref->isEqual($c) || 
			$c->FindChild ( $parent_ref ) ) )
			{
			unset( $parent_ref );
			}

		if ( $c->parent->uniq )
			{
             #  If it has a parent already, that's there because we saw this message
             #  in a references field, and presumed a parent based on the other
             #  entries in that field.  Now that we have the actual message, we can
             #  be more definitive, so throw away the old parent and use this new one.
             #  Find this container in the parent's child-list, and unlink it.
			 unset( $prev );
			for( $prev, $rest = &$c->parent->child;
					$rest->uniq;
					$prev=&$rest, $rest=&$rest->next )
				{
				if ( $rest->isEqual($c) ) { break;}
				}
			if ( !$rest->uniq )
				{
				print "ERROR: didnt find " . $c->uniq 
				       . " in child list " . $c->parent->uniq . "<br>"; 
				}

			if ( !$prev->uniq )
				{
				$c->parent->child = &$c->next;
				}
			else
				{
				$prev->next = &$c->next;
				}
			unset( $c->next );
			unset( $c->parent );
			}

		#  If we have a parent, link c into the parent's child list
		if ( $parent_ref->uniq )
			{
			$c->parent = &$parent_ref;
			$c->next =  &$parent_ref->child;
			$parent_ref->child = &$c;
			}
		}

	#  Find the root set of the ThreadContainers (and return a root node.)  
	$root = &newThreadContainer ();
	reset ( $id_table );
	while ( list( $k, $v ) = each( $id_table ) )
		{
		$scanned++;
		$c = &$id_table[ $k ]; 
		if ( !$c->parent->uniq )
			{
			if ( $c->next->uniq )
				{
				print "ERROR: " . $c->uniq . "has no parent but next<br>";
				}
			$c->next = &$root->child;
			$root->child = &$c;
			$inrootset++;
			}
		}

	pruneEmptyContainers ( $root );
#	reverseChildren ( $root );
	gatherSubjects( $root );
	unset ($msg_table);
	unset ($id_table);
	$nlist=array();
	$thread=$root->child->flushThread( $nlist );
	if ($config["tdbg"]) timedebug_stop ($tdbg);

	if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: making graphs");
	for ( $i=count($thread)-1; $i>=0; $i-- )
		{
		if ( $thread[$i+1]->graph[$thread[$i]->nesting-1] )
			{
			if ( $thread[$i+1]->threadid == $thread[$i-1]->threadid )
				{
				$thread[$i]->graph[$thread[$i]->nesting-1]="c";
				}
			elseif ( $thread[$i+1]->threadid == $thread[$i]->threadid ) 
				{
				$thread[$i]->graph[$thread[$i]->nesting-1]="n";
				}
			else
				{
				$thread[$i]->graph[$thread[$i]->nesting-1]="t";
				}
			}
		else
			{
			$thread[$i]->graph[$thread[$i]->nesting-1]="t";
			}
		for ( $k=$i-1; 
				$k>=0 &&
				$thread[$k]->threadid == $thread[$i]->threadid && 
				$thread[$k]->nesting > $thread[$i]->nesting; 
				$k-- )
			{
			$thread[$k]->graph[$thread[$i]->nesting-1]="p";
			}
		}
	
	$list=$nlist;
	if ($config["tdbg"]) timedebug_stop ($tdbg);

	if ( $config["cachethreshold"] >0 && $config["cachethreshold"] < $Nmsgs )
		{
		if ($config["tdbg"]) $tdbg=timedebug_start ("GetThreadedList: save to DISK cache");
		$cache=serialize (array ("list"=>$list, "thread"=>$thread));

		$cname = tempnam( GetCacheDir(), "");
		$fp = fopen ( $cname, "w");
		fputs ( $fp, $cache );
		fclose ($fp);

		$query="INSERT INTO " . $dbconfig["threadcache"] . " (fingerprint, Nmsgs, mailbox, username, type, cache) VALUES (" . 
				"'" . $fingerprint_thread . "'," . 
				"'" . $Nmsgs . "'," . 
				"'" . SQLString( $mbx ) . "'," .
				"'" . $login["username"] . "'," .
				"'T'," .
				"'" . SQLString( $cname ). "')";
		dbQuery ( $query );
		if ($config["tdbg"]) timedebug_stop ($tdbg);
		}
	}
?>
